# vim:ts=4:sw=4:expandtab
__author__ = "Carlos Descalzi"

from abc import ABCMeta, abstractmethod
from typing import Any, Mapping, Type
import copy


class Handler(metaclass=ABCMeta):
    """
    Interface for custom serialization handlers
    """

    @abstractmethod
    def encode(self, value: Any) -> Any:
        """
        Returns a json-friendly representation of the given value
        """
        return value

    @abstractmethod
    def decode(self, value_dict: Any) -> Any:
        """
        Converts back the dictionary into a desired type.
        """
        return value_dict


class CollectionOf:
    def __init__(self, type: Type):
        self._type = type

    @property
    def type(self) -> Type:
        return self._type


class ArrayOf(CollectionOf):
    """
    Describe an array of a given type.
    """

    pass


class MapOf(CollectionOf):
    """
    Describe a map a given value type.
    """

    pass


class Property(object):
    """
    Defines a property for a JSON object
    It allows to customize the json field name and the type
    used to deserialize the object.
    """

    def __init__(
        self,
        json=None,
        type=None,
        default=None,
        handler=None,
        enum_as_str=False,
        allow_none=True,
        validator=None,
        title=None,
        description=None,
        group=None,
    ):
        """
        Parameters:
        json: The json field name, can be different from the field name.
        type: The value type, useful when serializing data with no type information
        default: The default value when an object is instantiated. If it is a function/lambda
            it will invoke it to get a new value, otherwise this value is copied for each instance
        handler: Custom serialization handler for the value contained in the property 
        enum_as_str: Handle enumerators as strings.
        allow_none: Allows None as value.
        validator: Allow specify a validator for this property. Possible values are, None for no validation, 
            an implementation of Validator interface, or "default" to use default validation.
        title: A title to be placed on json schema.
        description: A description to be placed in json schema.
        group: can be a string or a list of strings. Can be used later to filter out what fields will be dumped.
        """
        self._json = json
        self._type = type
        self._default = default
        self._handler = handler
        self._enum_as_str = enum_as_str
        self._allow_none = allow_none
        self._validator = validator
        self._title = title
        self._description = description
        self._group = group

    @property
    def json(self) -> str:
        """
        JSON Field name, by default is the name of the attribute
        """
        return self._json

    @property
    def type(self):
        """
        Vault type, only required when value is instance of JsonObject
        """
        return self._type

    @property
    def default(self) -> Any:
        """
        Default value when object is instantiated.
        If it is a callable, it will be called to get the new value,
        otherwise this value will be used as prototype and copied on each instance.
        """
        return self._default

    @property
    def handler(self) -> Handler:
        """
        Custom handler for serializing/deserializing this field value.
        """
        return self._handler

    @property
    def enum_as_str(self) -> bool:
        """
        Determines if enums must be handled as string or int
        """
        return self._enum_as_str

    @property
    def validator(self):
        return self._validator

    @property
    def allow_none(self):
        return self._allow_none

    def default_val(self) -> Any:
        """
        Returns a new instance of the default vault for this field.
        """
        default = self.default
        if callable(default):
            return default()
        elif default is not None:
            return copy.deepcopy(default)
        return None

    @property
    def title(self):
        return self._title

    @property
    def description(self):
        return self._description

    @property
    def group(self):
        return self._group



# vim:ts=4:sw=4:expandtab
import unittest
from podm import JsonObject, Property, Handler, Processor, ArrayOf, MapOf, add_alias
from collections import OrderedDict
from datetime import datetime
from .common import (
    Entity,
    Company,
    Sector,
    Employee,
    DateTimeHandler,
    TestObject,
    TestObject2,
    Child,
    Parent,
)


class TestAliasedObject(JsonObject):
    field1 = Property()


class TestSerialization(unittest.TestCase):
    def test_properties(self):
        self.assertEqual(set(Entity.property_names()), set(["oid", "created"]))
        self.assertEqual(
            set(Company.property_names()), set(["oid", "created", "description", "company_name"]),
        )
        self.assertEqual(
            set(Company.json_field_names()), set(["oid", "created", "description", "company-name"]),
        )

    def test_accessors(self):
        company = Company()
        company.oid = "123123123"
        company.company_name = "master"
        company.description = "master company"

        self.assertEqual("master", company.company_name)

        self.assertEqual("master", company.get_company_name())
        self.assertEqual("master company", company.description)
        self.assertTrue(company._used_getter)
        company._used_getter = False
        self.assertEqual("master company", company.get_description())
        self.assertTrue(company._used_getter)

        company.set_description("desc 2")
        self.assertEqual("desc 2", company.description)
        self.assertEqual("desc 2", company._description)

    def test_to_dict(self):
        company = Company()
        company.company_name = "master"
        company.description = "master company"

        data = company.to_dict()

        self.assertTrue("py/object" in data)
        print(data["py/object"])
        self.assertTrue("oid" in data)
        self.assertTrue("created" in data)
        self.assertIsNotNone(data["created"])
        self.assertTrue("company-name" in data)
        self.assertEqual("master", data["company-name"])
        self.assertTrue("description" in data)
        self.assertEqual("master company", data["description"])

    def test_deserialize_no_module(self):
        data = {
            "py/object": "Company",
            "py/state": {"company-name": "master", "description": "some description"},
        }
        company = JsonObject.parse(data, __name__)
        self.assertTrue(isinstance(company, Company))
        self.assertEqual("master", company.company_name)
        self.assertEqual("some description", company.description)

    def test_deserialize_with_module(self):
        data = {
            "py/object": "test.common.Company",
            "py/state": {"company-name": "master", "description": "some description"},
        }
        company = JsonObject.parse(data)
        self.assertTrue(isinstance(company, Company))
        self.assertEqual("master", company.company_name)
        self.assertEqual("some description", company.description)

    def test_deserialize_3(self):
        data = {"company-name": "master", "description": "some description"}
        company = Company.from_dict(data)
        self.assertTrue(isinstance(company, Company))
        self.assertEqual("master", company.company_name)
        self.assertEqual("some description", company.description)

    def test_kwargs_constructor(self):
        company = Company(company_name="master", description="some description")

        # check accessors
        self.assertEqual("master", company.company_name)
        self.assertEqual("some description", company.description)

        # check internal state
        self.assertEqual("master", company._company_name)
        self.assertEqual("some description", company._description)

    def test_collections(self):
        sector = Sector()
        employee = Employee(oid="1234", name="xx1")

        sector.employees.append(employee)

        data = sector.to_dict()

        self.assertTrue(isinstance(data["employees"], list))

        for employee in data["employees"]:
            self.assertEqual("1234", employee["oid"])
            self.assertEqual("xx1", employee["name"])

        new_sector = Sector.from_dict(data)

        self.assertTrue(isinstance(new_sector.employees, list))
        self.assertEqual(1, len(new_sector.employees))

        self.assertTrue(isinstance(new_sector.employees[0], Employee))
        self.assertEqual("1234", new_sector.employees[0].oid)
        self.assertEqual("xx1", new_sector.employees[0].name)

    def test_default(self):
        k1 = Company()
        k2 = Company()

        self.assertIsNotNone(k1.created)
        self.assertIsNotNone(k2.created)

        self.assertNotEqual(k1.created, k2.created)

    def test_custom_objects(self):

        company = Company(created=datetime.now())

        company_dict = company.to_dict()

        self.assertTrue("created" in company_dict)
        self.assertTrue(isinstance(company_dict["created"], datetime))

    def test_handler(self):

        obj = TestObject()

        obj_dict = obj.to_dict()
        self.assertTrue(isinstance(obj_dict["date-time"], dict))
        self.assertTrue("year" in obj_dict["date-time"])

        obj2 = TestObject.from_dict(obj_dict)
        self.assertTrue(isinstance(obj2.date_time, datetime))

        self.assertTrue(obj2._deserialized)

    def test_handler_2(self):
        class BoolHandler(Handler):
            def encode(self, val):
                return val

            def decode(self, val):
                print("decode", val)
                return str(val).lower() == "true" if val is not None else None

        bool_handler = BoolHandler()

        class TestObject2(JsonObject):
            some_boolean_1 = Property(handler=bool_handler)
            some_boolean_2 = Property(handler=bool_handler)
            some_boolean_3 = Property(handler=bool_handler)

        obj_dict = {
            "some_boolean_1": "True",
            "some_boolean_2": "false",
            "some_boolean_3": None,
        }

        obj = TestObject2.from_dict(obj_dict)

        self.assertIsInstance(obj.some_boolean_1, bool)
        self.assertTrue(isinstance(obj.some_boolean_2, bool))
        self.assertIsNone(obj.some_boolean_3)

    def test_str(self):
        class TestObject3(JsonObject):
            val1 = Property()
            val2 = Property()
            val3 = Property()

        obj_dict = {"val1": "Hi", "val2": True, "val3": None}

        obj = TestObject3.from_dict(obj_dict)

        obj_str = str(obj)

        self.assertEqual("TestObject3:val1=Hi;val2=True;val3=None", obj_str)

    def test_ordered_dict(self):
        class TestObject4(JsonObject):
            val1 = Property()

        obj = TestObject4()
        obj.val1 = OrderedDict(key1="value1", key2="value2", key3="value3", key4="value4")

        serialized = obj.to_dict(OrderedDict)
        self.assertTrue(isinstance(serialized, OrderedDict))
        self.assertTrue(isinstance(serialized["val1"], OrderedDict))

        serialized = obj.to_dict()
        self.assertFalse(isinstance(serialized, OrderedDict))
        self.assertTrue(isinstance(serialized["val1"], OrderedDict))

    def test_yaml(self):
        class TestObject4(JsonObject):
            val1 = Property()
            val2 = Property()

        obj = TestObject4()
        obj.val1 = {"key1": "val1"}
        obj.val2 = [1, 2, 3]
        import yaml

        print()
        print(yaml.dump(obj.to_dict()))

    def test_enums_str(self):
        from enum import Enum, auto

        class Enum1(Enum):
            VAL1 = 1
            VAL2 = 2

        class TestObject(JsonObject):
            val = Property(type=Enum1, enum_as_str=True)

        obj = TestObject()
        obj.val = Enum1.VAL1

        data = obj.to_dict()
        self.assertEqual("VAL1", data["val"])

        obj2 = TestObject.from_dict(data)
        self.assertEqual(Enum1.VAL1, obj2.val)

    def test_enums_int(self):
        from enum import Enum, auto

        class Enum1(Enum):
            VAL1 = 1
            VAL2 = 2

        class TestObject(JsonObject):
            val = Property(type=Enum1)

        obj = TestObject()
        obj.val = Enum1.VAL1

        data = obj.to_dict()
        self.assertEqual(1, data["val"])

        obj2 = TestObject.from_dict(data)
        self.assertEqual(Enum1.VAL1, obj2.val)

    def test_int_enums_int(self):
        from enum import IntEnum, auto

        class Enum1(IntEnum):
            VAL1 = 1
            VAL2 = 2

        class TestObject(JsonObject):
            val = Property(type=Enum1)

        obj = TestObject()
        obj.val = Enum1.VAL1

        data = obj.to_dict()
        self.assertEqual(1, data["val"])
        self.assertFalse(isinstance(data["val"], IntEnum))

        obj2 = TestObject.from_dict(data)
        self.assertEqual(Enum1.VAL1, obj2.val)

    def test_plain_props(self):
        class TestObject(JsonObject):
            prop1 = Property()

            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                self._prop2 = None

            def set_prop2(self, prop2):
                self._prop2 = prop2

            def get_prop2(self):
                return self._prop2

            prop2 = property(get_prop2, set_prop2)

        obj = TestObject()

        self.assertIsNone(obj.prop2)
        try:
            obj.prop2 = 1
        except AttributeError:
            self.assertTrue(False)

        val = obj.prop2

        self.assertEqual(1, val)

    def test_jsonpickle_format(self):
        class TestObject(JsonObject):

            __jsonpickle_format__ = True

            prop1 = Property()

        obj = TestObject()
        obj.prop1 = "hello"

        data = obj.to_dict()

        self.assertIn("py/object", data)
        self.assertIn("py/state", data)

        self.assertIn("prop1", data["py/state"])
        self.assertEqual("hello", data["py/state"]["prop1"])

    def test_nested(self):

        test_obj = TestObject2()
        test_obj.property1 = "hello"

        data = {"obj": test_obj.to_dict()}

        parsed = JsonObject.parse(data)

        self.assertIsInstance(parsed, dict)
        self.assertIn("obj", parsed)
        self.assertIsInstance(parsed["obj"], TestObject2)

    def test_explicity_typing_in_array(self):

        data = {"children": [{"property1": "value1"}]}

        obj = Parent.from_dict(data)

        self.assertIsNotNone(obj)
        self.assertEqual(1, len(obj.children))
        self.assertIsInstance(obj.children[0], Child)

    def test_group_filter(self):
        class TestObject5(JsonObject):
            field1 = Property(group="group1")
            field2 = Property(group="group2")
            field3 = Property(group="group2")

        obj = TestObject5(field1=1, field2=2, field=3)
        result = obj.to_dict(group_filter="group1")
        self.assertEqual(2, len(result))
        self.assertIn("field1", result)
        self.assertNotIn("field2", result)
        self.assertNotIn("field3", result)

        result = obj.to_dict(group_filter="group2")
        self.assertEqual(3, len(result))
        self.assertNotIn("field1", result)
        self.assertIn("field2", result)
        self.assertIn("field3", result)

    def test_alias(self):
        add_alias("test_object_1", TestAliasedObject.object_type_name())

        obj = TestAliasedObject(field1="x")
        data = obj.to_dict()

        self.assertEqual("test_object_1", data["py/object"])

        deserialized = JsonObject.parse(data)

        self.assertIsInstance(deserialized, TestAliasedObject)


if __name__ == "__main__":
    unittest.main()



# vim:ts=4:sw=4:expandtab
__author__ = "Carlos Descalzi"

from abc import ABCMeta, abstractmethod
from enum import Enum, IntEnum
from .meta import Handler
from typing import Any, Type, Mapping


class PropertyHandler(metaclass=ABCMeta):
    def __init__(self, obj_type, name, definition):
        self._obj_type = obj_type
        self._name = name
        self._definition = definition

    def name(self):
        return self._name

    def handler(self) -> Handler:
        """
        Returns a custom handler for the property value, or None.
        """
        return None

    def field_type(self) -> Type:
        """
        Returns the type of the field value, intended when such value
        wants to be deserialized to a BaseJsonObject, and dictionary
        has no type information.
        """
        return None

    @abstractmethod
    def set(self, target: Any, value: Any):
        pass

    @abstractmethod
    def get(self, target: Any) -> Any:
        pass

    @abstractmethod
    def json(self) -> str:
        """
        Return the actual json name
        """
        pass

    @abstractmethod
    def field_name(self) -> str:
        """
        Return the field name.
        """
        pass

    def json_field_type(self, type_definitions={}):
        """
        Converts actual field type into javascript types.
        Subclasses may extend it
        """
        field_type = self.field_type()

        if field_type is None:
            return "object"
        if field_type == str:
            return "string"
        elif field_type == bool:
            return "boolean"
        elif field_type in [int, float]:
            return "number"
        elif field_type == list:
            return "array"
        elif issubclass(field_type, Enum):
            return "number"
        elif field_type.__name__ in type_definitions:
            return {"$ref": f"#/definitions/{field_type.__name__}"}
        else:
            return "object"

    def schema(self, type_definitions={}) -> Mapping:
        """
        Returns the json schema definition of the property
        """
        return {"type": self.json_field_type(type_definitions)}

    def validator(self):
        return None

    def allow_none(self):
        return True

    @property
    def group(self):
        return None


class RichPropertyHandler(PropertyHandler):
    """
    Adds features like default property value, enumerator handling and 
    better property accessor usage
    """

    @abstractmethod
    def init(self, target, value):
        """
        Initializes a property with a given value, or 
        a default value provided by this implementation.
        """
        pass

    @abstractmethod
    def enum_as_str(self) -> bool:
        """
        Boolean flag that determines if flags are stored/retrieved
        as string literals or integers.
        """
        pass

    @abstractmethod
    def getter(self):
        pass

    @abstractmethod
    def getter_name(self) -> str:
        pass

    @abstractmethod
    def setter(self):
        pass

    @abstractmethod
    def setter_name(self) -> str:
        pass


class DefaultSetter:
    def __init__(self, field_name):
        self._field_name = field_name

    def __call__(self, target, value):
        target.__dict__[self._field_name] = value


class DefaultGetter:
    def __init__(self, field_name):
        self._field_name = field_name

    def __call__(self, target):
        return target.__dict__[self._field_name]


class DefaultPropertyHandler(RichPropertyHandler):
    def __init__(self, obj_type, name, definition):
        super().__init__(obj_type, name, definition)
        self._field_name = "_%s" % name
        self._setter_name = "set_%s" % name
        self._getter_name = "get_%s" % name

        getter = obj_type.__dict__.get(self._getter_name)
        self._getter = getter or DefaultGetter(self._field_name)

        setter = obj_type.__dict__.get(self._setter_name)
        self._setter = setter or DefaultSetter(self._field_name)

    def init(self, target, value):
        self.set(target, value or self._definition.default_val())

    def set(self, target, value):
        self._setter(target, value)

    def get(self, target):
        return self._getter(target)

    def json(self):
        return self._definition.json or self._name

    def handler(self):
        return self._definition.handler

    def field_name(self):
        return self._field_name

    def field_type(self):
        return self._definition.type

    def enum_as_str(self):
        return self._definition.enum_as_str

    def getter(self):
        return self._getter

    def getter_name(self):
        return self._getter_name

    def setter(self):
        return self._setter

    def setter_name(self):
        return self._setter_name

    def validator(self):
        return self._definition.validator

    def allow_none(self):
        return self._definition.allow_none

    @property
    def group(self):
        return self._definition.group

    def schema(self, type_definitions={}):
        schema = super().schema(type_definitions)

        default = self._definition.default
        if default is not None and not callable(default):
            schema["default"] = self._definition.default

        if self._definition.title:
            schema["title"] = self._definition.title

        if self._definition.description:
            schema["description"] = self._definition.description

        if self.field_type() and issubclass(self.field_type(), Enum) and self.enum_as_str:
            schema["type"] = "string"
            schema["enum"] = self.field_type()._member_names_

        return schema



from unittest import TestCase
from podm import JsonObject, Property, ArrayOf
from podm.util.mongo import QueryHelper


class Customer(JsonObject):
    first_name = Property(json="first-name")
    last_name = Property(json="last-name")
    tax_id = Property(json="tax-id")


class Item(JsonObject):
    product_id = Property(json="product-id")
    quantity = Property()
    unit_price = Property("unit-price")


class Invoice(JsonObject):
    customer = Property(type=Customer)
    items = Property(type=ArrayOf(Item))


class JsonPickleInvoice(JsonObject):
    __jsonpickle_format__ = True
    customer = Property(type=Customer)
    items = Property(type=ArrayOf(Item))


class TestQueryHelper(TestCase):
    def test_basic_operators(self):
        qhi = QueryHelper(Invoice)

        self.assertEqual({"items.product-id": 10}, (qhi.items.product_id == 10).expr())
        self.assertEqual({"items.product-id": {"$ne": 10}}, (qhi.items.product_id != 10).expr())
        self.assertEqual({"items.product-id": {"$lt": 10}}, (qhi.items.product_id < 10).expr())
        self.assertEqual({"items.product-id": {"$lte": 10}}, (qhi.items.product_id <= 10).expr())
        self.assertEqual({"items.product-id": {"$gt": 10}}, (qhi.items.product_id > 10).expr())
        self.assertEqual({"items.product-id": {"$gte": 10}}, (qhi.items.product_id >= 10).expr())
        self.assertEqual({"items.product-id": {"$in": [1, 2]}}, qhi.items.product_id.in_([1, 2]).expr())
        self.assertEqual({"items.product-id": {"$nin": [1, 2]}}, qhi.items.product_id.nin([1, 2]).expr())
        self.assertEqual({"items.product-id": {"$exists": True}}, (qhi.items.product_id.exists()).expr())

    def test_logic_operators(self):
        qhi = QueryHelper(Invoice)
        self.assertEqual(
            {"$and": [{"items.product-id": 10}, {"items.quantity": 5}]},
            qhi.and_(qhi.items.product_id == 10, qhi.items.quantity == 5).expr(),
        )
        self.assertEqual(
            {"items.product-id": 10, "items.quantity": 5},
            qhi.join(qhi.items.product_id == 10, qhi.items.quantity == 5).expr(),
        )
        self.assertEqual(
            {"$or": [{"items.product-id": 10}, {"items.quantity": 5}]},
            qhi.or_(qhi.items.product_id == 10, qhi.items.quantity == 5).expr(),
        )
        self.assertEqual(
            {"$nor": [{"items.product-id": 10}, {"items.quantity": 5}]},
            qhi.nor(qhi.items.product_id == 10, qhi.items.quantity == 5).expr(),
        )
        self.assertEqual(
            {"$not": {"items.product-id": 10}}, qhi.not_(qhi.items.product_id == 10).expr(),
        )

        self.assertEqual(
            {"$not": {"$and": [{"items.product-id": 10}, {"items.quantity": 5}]}},
            qhi.not_(qhi.and_(qhi.items.product_id == 10, qhi.items.quantity == 5)).expr(),
        )

    def test_array_ref(self):
        qhi = QueryHelper(Invoice)
        self.assertEqual({"items[0].product-id": 10}, (qhi.items[0].product_id == 10).expr())

    def test_json_pickle_format(self):
        qhi = QueryHelper(JsonPickleInvoice)

        self.assertEqual({"py/state.items.product-id": 10}, (qhi.items.product_id == 10).expr())

    def test_obj_type_expr(self):

        qhi = QueryHelper(Invoice)

        self.assertEqual({"py/object": "test.test_query_helper.Invoice"}, qhi.is_type_expr().expr())
